@@ -525,9 +525,9 @@ module Vars = struct
 
     let get_int_opt t var =
       Option.bind (get_opt t var) ~f:(fun s ->
-          match int_of_string s with
-          | x -> Some x
-          | exception _ -> fail "Value of %S is not an integer: %s." var s)
+          match Int.of_string s with
+          | Some x -> Some x
+          | None -> fail "Value of %S is not an integer: %s." var s)
 
     let get_words t var =
       match get_opt t var with
